from option_merge.versioning import VersionedDict
from option_merge.merge import MergedOptions
from option_merge.joiner import dot_joiner

def prefixed_path_list(path, prefix=None):
    """Return the prefixed version of this path as a list"""
    res_type = type(path)
    if prefix:
        res = prefix + path
    else:
        res = list(path)
    return res, dot_joiner(res, res_type)

def prefixed_path_string(path, prefix=""):
    """Return the prefixed version of this string"""
    while path and path[0] == '.':
        path = path[1:]

    while path and path[-1] == ".":
        path = path[:-1]

    while prefix and prefix[0] == ".":
        prefix = prefix[1:]

    while prefix and prefix[-1] == ".":
        prefix = prefix[:-1]

    if not prefix:
        return path, path
    elif not path:
        return prefix, prefix
    else:
        res = "{0}.{1}".format(prefix, path)
        return res, res

def make_dict(first, rest, data):
    """Make a dictionary from a list of keys"""
    last = first
    result = VersionedDict({first: data})
    current = result
    for part in rest:
        current[last] = VersionedDict({})
        current = current[last]
        current[part] = data
        last = part

    return result

def merge_into_dict(target, source, seen=None, ignore=None):
    """Merge source into target"""
    if ignore is None:
        ignore = []

    if hasattr(source, "is_dict") and source.is_dict:
        if hasattr(source, "as_dict"):
            source = source.as_dict(seen=seen, ignore=ignore)
        else:
            source = dict(source)

    for key in source.keys():
        if key in ignore:
            continue
        val = source[key]

        is_dict = lambda item: type(item) in (dict, VersionedDict, MergedOptions) or isinstance(item, dict)
        if is_dict(val):
            if not is_dict(target.get(key)):
                target[key] = {}
            merge_into_dict(target[key], val, seen=seen, ignore=ignore)
        else:
            target[key] = val



"""
Option_merge uses this Path object to encapsulate the idea of a path and the
converters that are available to use.

We are able to use this to store a reference to the root of the configuration
as well as whether the converters should be ignored or not.

It's purpose is to behave like a string regardless of whether it is a string
or a list of strings.
"""

from option_merge.joiner import dot_joiner, join
from option_merge.not_found import NotFound

import six

class Path(object):
    """
    Encapsulate a path; a root configuration; a list of converters; and whether
    the converters should be used or not

    A path may be just a string or a list of strings.
    """
    @classmethod
    def convert(kls, path, configuration=None, converters=None, ignore_converters=None, joined=None):
        """
        Get us a Path object from this path

        If path is already a Path instance, it is returned as is.

        Otherwise, a joined version of the string is created and used,
        along with the other kwargs to this function, to produce a new Path instance
        """
        path_type = type(path)
        if path_type is Path:
            return path
        else:
            joined = dot_joiner(path, item_type=path_type)
            return Path(path, configuration, converters, ignore_converters, joined=joined)

    def __init__(self, path, configuration=None, converters=None, ignore_converters=False, joined=None, joined_function=None):
        self.path = path
        self.path_type = type(self.path)
        self.path_is_string = self.path_type in (str, ) + six.string_types

        self._joined = joined
        self._joined_function = joined_function

        self.converters = converters
        self.configuration = configuration
        self.ignore_converters = ignore_converters

    def __unicode__(self):
        """alias for self.joined"""
        return self.joined()

    def __str__(self):
        """alias for self.joined"""
        return self.joined()

    def __nonzero__(self):
        """Whether we have any path or not"""
        return any(self.path)

    def __len__(self):
        """
        The length of our path

        * If we have no path, then 0
        * if path is a string, then 1
        * if path is an array, then the length of the array
        """
        if self.path_is_string:
            if self.path:
                return 1
            else:
                return 0
        else:
            if self.path_type in (list, tuple):
                if not any(item for item in self.path):
                    return 0
            return len(self.path)

    def __iter__(self):
        """Iterate through the parts of our path"""
        if self.path_is_string:
            if self.path:
                yield self.path
        else:
            for part in self.path:
                yield part

    def __repr__(self):
        return "<Path({0})>".format(str(self))

    def __eq__(self, other):
        """
        Compare the joined version of this path
        and the joined version of the other path
        """
        joined = self.joined()
        if not other and not joined:
            return True

        if other and joined:
            return dot_joiner(other) == self.joined()

        return False

    def __ne__(self, other):
        """Negation of whether other is equal to this path"""
        return not self.__eq__(other)

    def __add__(self, other):
        """Create a copy of this path joined with other"""
        if not other:
            return self.clone()
        else:
            return self.using(join(self, other))

    def __hash__(self):
        """The hash of the joined version of this path"""
        return hash(self.joined())

    def __getitem__(self, key):
        """
        If the path is a string, treat it as a list of that one string,
        otherwise, treat path as it is
        and get the index of the path as specified by key
        """
        path = self.path
        if self.path_is_string:
            path = [path]
        return path[key]

    def without(self, base):
        """Return a clone of this path without the base"""
        base_type = type(base)
        if base_type not in (str, ) + six.string_types:
            base = dot_joiner(base, base_type)

        if not self.startswith(base):
            raise NotFound()

        if self.path_is_string:
            path = self.path[len(base):]
            while path and path[0] == ".":
                path = path[1:]
            return self.using(path, joined=path)
        else:
            if not base:
                res = [part for part in self.path]
            else:
                res = []
                for part in self.path:
                    if not base:
                        res.append(part)
                        continue

                    part_type = type(part)
                    if part_type in (str, ) + six.string_types:
                        joined_part = part
                    else:
                        joined_part = dot_joiner(part, part_type)

                    if base.startswith(joined_part):
                        base = base[len(joined_part):]
                        while base and base[0] == ".":
                            base = base[1:]
                    elif joined_part.startswith(base):
                        res.append(joined_part[len(base):])
                        base = ""

            return self.using(res, joined=dot_joiner(res, list))

    def prefixed(self, prefix):
        """Return a clone with this prefix to the path"""
        if not prefix:
            return self.clone()
        else:
            return self.using(join(prefix, self))

    def first_part_is(self, key):
        """Return whether the first part of this path is this string"""
        if self.path_is_string:
            return self.path.startswith(str(key) + '.')
        if not self.path:
            return not bool(key)
        if self.path_type is list:
            return self.path[0] == key
        if self.path_type is Path:
            return self.path.first_part_is(key)
        return self.joined().startswith(str(key) + '.')

    def startswith(self, base):
        """Does the path start with this string?"""
        if self.path_is_string:
            return self.path.startswith(base)
        if not self.path:
            return not bool(base)
        if self.path_type is list and len(self.path) is 1:
            return self.path[0].startswith(base)
        return self.joined().startswith(base)

    def endswith(self, suffix):
        """Does the path end with this string?"""
        return self.joined().endswith(suffix)

    def using(self, path, configuration=None, converters=None, ignore_converters=False, joined=None):
        """Return a clone of this path and override with provided values"""
        if configuration is None:
            configuration = self.configuration
        if converters is None:
            converters = self.converters

        if path == self.path and self.configuration is configuration and self.converters is converters and self.ignore_converters is ignore_converters:
            return self

        joined_function = None
        if joined is None:
            if type(path) is Path:
                joined_function = lambda: dot_joiner(path.path, path.path_type)
            else:
                joined_function = lambda: dot_joiner(path)
        return self.__class__(path, configuration, converters, ignore_converters=ignore_converters, joined_function=joined_function)

    def clone(self):
        """Return a clone of this path with all the same values"""
        joined_function = lambda: dot_joiner(self.path, self.path_type)
        return self.__class__(self.path, self.configuration, self.converters, self.ignore_converters, joined_function=joined_function)

    def ignoring_converters(self, ignore_converters=True):
        """Return a clone of this path with ignore_converters set to True"""
        if self.ignore_converters == ignore_converters:
            return self
        return self.using(self.path, ignore_converters=ignore_converters, joined=self.joined())

    def do_conversion(self, value):
        """
        Do the conversion on some path if any conversion exists

        Return (converted, did_conversion)

        Where ``did_conversion`` is a boolean indicating whether a conversion
        took place.
        """
        converter, found = self.find_converter()
        if not found:
            return value, False
        else:
            converted = converter(self, value)
            self.converters.done(self, converted)
            if hasattr(converted, "post_setup"):
                converted.post_setup()
            return converted, True

    def find_converter(self):
        """Find appropriate converter for this path"""
        if self.ignore_converters:
            return None, False
        return self.converters.matches(self)

    def converted(self):
        """Determine if this path has been converted"""
        if self.converters:
            return self.converters.converted(self)
        return False

    def converted_val(self):
        """Return the converted value for this path"""
        return self.converters.converted_val(self)

    def waiting(self):
        """Return whether we're waiting for this value"""
        return self.converters.waiting(self)

    def joined(self):
        """Return the dot_join of of the path"""
        joined = self._joined
        if self._joined is None and self._joined_function is not None:
            joined = self._joined = self._joined_function()

        if joined is None:
            if self.path_is_string:
                joined = self._joined = self.path
            else:
                joined = self._joined = dot_joiner(self.path, self.path_type)
        return joined




"""
This is the main class and the entry point for the programmer.

It provides a mechanism to treat multiple dictionaries as if they were one
dictionary.

Also provided is the ability to reference parts and still maintain the idea of it
being one dictionary.

With the ability to delete from the dictionary and the ability to convert values
on access.
"""

from option_merge.versioning import versioned_iterable, versioned_value, VersionedDict
from option_merge.converter import Converters, Converter
from option_merge.not_found import NotFound
from option_merge.joiner import dot_joiner
from option_merge.path import Path

from collections import Mapping
import logging
import six

log = logging.getLogger("option_merge.merge")

class KeyValuePairsConverter(object):
    """Converts a list of key,value pairs to a dictionary"""
    def __init__(self, pairs, source=None):
        self.pairs = pairs
        self.source = source

    def convert(self):
        """Return us a MergedOptions from our pairs"""
        from option_merge import helper as hp
        return MergedOptions().using(*[hp.make_dict(key[0], key[1:], value) for key, value in self.pairs], source=self.source)

class AttributesConverter(object):
    """Converts an object with particular attributes to a dictionary"""
    def __init__(self, obj, attributes=None, include_underlined=False, lift=None, ignoreable_values=None, source=None):
        self.obj = obj
        self.lift = lift
        self.source = source
        self.attributes = attributes
        self.ignoreable_values = ignoreable_values
        self.include_underlined = include_underlined

        if isinstance(self.lift, six.string_types):
            self.lift = [self.lift]

    def convert(self):
        """Return us a MergedOptions from our attributes"""
        options = MergedOptions()
        attributes = self.attributes
        if not attributes:
            attributes = dir(self.obj)
            if not self.include_underlined:
                attributes = [attr for attr in attributes if not attr.startswith("_")]

        for attr in attributes:
            if hasattr(self.obj, attr):
                val = getattr(self.obj, attr)
                if not self.ignoreable_values or val not in self.ignoreable_values:
                    options[attr] = val

        if self.lift:
            lifted = MergedOptions()
            lifted.storage.add(Path.convert(self.lift), options)
            return lifted
        else:
            return options

class ConverterProperty(object):
    """Creates a Property for accessing a converter"""
    def __init__(self, converter):
        self.converter = converter

    def __get__(self, obj=None, owner=None):
        return lambda *args, **kwargs: self.converter(*args, **kwargs).convert()

class MergedOptions(dict, Mapping):
    """
    Wrapper around multiple dictionaries to behave as one.

    Usage::

        options = MergedOptions.using(options1, options2, source="SomePlace")

    Is equivalent to::

        options = MergedOptions()
        options.update(options1, source="SomePlace")
        options.update(options2, source="SomePlace")

    The later an option is added, the more influence it has.
    i.e. when a key is accessed, later options are looked at first.

    When you delete a key, it removes it from the first dictionary it can find.
    This means a key can change value when deleted rather than disappearing altogether

    It will also merge deeply.

    So::

        options1 = {'a':{'b':1, 'c':3}, 'b':5}
        options2 = {'a':{'b':4'}, 'd':7}
        merged = MergedOptions.using(options1, options2)
        merged['a'] == MergedOptions(prefix='a', <same_options>)
        merged['a']['b'] == 4
        merged['a']['c'] == 3
        merged['d'] == 7
    """

    Attributes = ConverterProperty(AttributesConverter)
    KeyValuePairs = ConverterProperty(KeyValuePairsConverter)

    def __init__(self, prefix=None, storage=None, dont_prefix=None, converters=None, ignore_converters=False):
        self.prefix_list = prefix
        self.converters = converters
        self.dont_prefix = dont_prefix
        self.ignore_converters = ignore_converters
        if not self.dont_prefix:
            self.dont_prefix = []
        if not self.prefix_list:
            self.prefix_list = []
        if not self.converters:
            self.converters = Converters()
        if isinstance(self.prefix_list, six.string_types):
            self.prefix_list = [self.prefix_list]
        self.prefix_string = dot_joiner(self.prefix_list, list)

        self.storage = storage
        from option_merge.storage import Storage
        if self.storage is None:
            self.storage = Storage()

    @classmethod
    def using(cls, *options, **kwargs):
        """
        Convenience for calling update multiple times

        .. code-block:: python

            m = MergedOptions.using({"a": 1}, {"b": 2})

        is equivalent to:

        .. code-block:: python

            m = MergedOptions()
            m.update({"a": 1})
            m.update({"b": 2})

        Any kwargs given to ``using`` is passed into ``update`` for
        each provided dictionary.
        """
        prefix = kwargs.get('prefix')
        storage = kwargs.get('storage')
        converters = kwargs.get('converters')
        dont_prefix = kwargs.get('dont_prefix')
        ignore_converters = kwargs.get("ignore_converters")
        merged = cls(
              prefix=prefix, storage=storage, dont_prefix=dont_prefix
            , converters=converters, ignore_converters=ignore_converters
            )

        for opts in options:
            merged.update(opts, **kwargs)
        return merged

    @property
    def version(self):
        return self.storage.version

    def update(self, options, source=None, **kwargs):
        """
        Add new options to the storage under this prefix.

        The later options are added, the more influence they have.
        """
        if options is None: return
        self.storage.add(Path(self.prefix_list), options, source=source)

    @versioned_value
    def __getitem__(self, path, ignore_converters=False):
        """
        Access some path

        Return the first value it comes across

        Raise KeyError if nothing has the specified key

        for example:

        .. code-block:: python

            m = MergedOptions.using({"a": {"b": 2}})
            assert m['a.b'] == 2
        """
        path = self.converted_path(path, ignore_converters=ignore_converters or self.ignore_converters or getattr(path, "ignore_converters", False))
        for val, return_as_is in self.values_for(path, ignore_converters=path.ignore_converters):
            if return_as_is:
                return val

            if any(isinstance(val, unprefixed) for unprefixed in self.dont_prefix):
                return val
            elif type(val) in (dict, VersionedDict, MergedOptions) or isinstance(val, dict):
                return self.prefixed(path, already_prefixed=True)
            else:
                return val

        raise KeyError(path)

    def __contains__(self, path):
        """
        Ask storage if it has a path

        .. code-block:: python

            m = MergedOptions.using({"a": 1})

            assert "a" in m
            assert "b" not in m
        """
        try:
            self.storage.get(self.converted_path(path))
            return True
        except KeyError:
            return False

    def get(self, path, default=None, ignore_converters=False):
        """
        Get some path or return default value

        .. code-block:: python

            m = MergedOptions.using({"a": 1})

            assert m.get("a") == 1
            assert m.get("b") == None
            assert m.get("b", 2) == 2

        You may also specify ``ignore_converters`` and it won't take the
        the converters into account.

        .. code-block:: python

            m = MergedOptions.using({"a": 1})
            m.add_converter(Converter(convert=add_one, convert_path=["a"]))
            m.converters.activate()

            assert m["a"] == 2
            assert m.get("a", ignore_converters=True) == 1
        """
        try:
            return self.__getitem__(path, ignore_converters=ignore_converters)
        except KeyError:
            return default

    def source_for(self, path, chain=None):
        """
        Proxy self.storage.source_for

        Source is specifying in calls to ``__init__`` and ``update`` and this
        will find the entries for the specified path an return the first source
        it finds.
        """
        path = Path.convert(path, self).ignoring_converters(True)
        return self.storage.source_for(path, chain)

    def __setitem__(self, path, value):
        """
        Set a key in the storage

        This takes into account the prefix on this option as well as the provided
        path.

        .. code-block:: python

            m = MergedOptions.using({"a": 1})
            assertEqual(m.as_dict(), {"a": 1})

            a = m["a"]
            a['b'] = 2

            assertEqual(m.as_dict(), {"a": 1, "b": 2})
        """
        if isinstance(path, six.string_types):
            path = [path]
        self.storage.add(self.converted_path(path), value)

    def __delitem__(self, path):
        """
        Delete a key from the storage

        .. code-block:: python

            m = MergedOptions.using({"a": 1}, {"a": 2})
            assert m['a'] == 2

            del m['a']
            assert m['a'] == 1
        """
        self.storage.delete(self.converted_path(path))

    def __iter__(self):
        """Iterate over the keys"""
        return iter(self.keys())

    def __len__(self):
        """Get number of keys we have"""
        return len(list(self.keys()))

    def __eq__(self, other):
        """Equal to another merged options if has same storage and prefix"""
        return isinstance(other, self.__class__) and other.storage is self.storage and other.prefix_list == self.prefix_list

    @versioned_iterable
    def values_for(self, path, ignore_converters=False):
        """Get all known values for some path"""
        path = self.converted_path(path, ignore_converters=ignore_converters or getattr(path, "ignore_converters", False))
        if not path.ignore_converters and not path.waiting():
            if path.converted():
                yield path.converted_val(), True
                return

            if path.find_converter()[1]:
                untouched = self[path.ignoring_converters()]
                yield path.do_conversion(untouched)
                return

        for info in self.storage.get_info(path):
            try:
                yield info.value_after(path), False
            except NotFound:
                pass

    def prefixed(self, path, ignore_converters=False, already_prefixed=False):
        """Return a MergedOptions prefixed to this path"""
        return self.__class__(
              self.converted_path(path, ignore_converters=ignore_converters)
            , storage=self.storage
            , dont_prefix=self.dont_prefix
            , converters=self.converters
            , ignore_converters=ignore_converters
            )

    def root(self):
        """Return a MergedOptions looking at the root of the storage"""
        return self.__class__(
              ""
            , storage=self.storage
            , dont_prefix=self.dont_prefix
            , converters=self.converters
            , ignore_converters=self.ignore_converters
            )

    def wrapped(self):
        """
        Return a MergedOptions with this inside

        Equivalent to:

        .. code-block:: python

            m = MergedOptions.using("a")
            wrapped = MergedOptions.using(m, converters=m.converters, dont_prefix=m.dont_prefix)
        """
        return self.__class__.using(self, converters=self.converters, dont_prefix=self.dont_prefix)

    def keys(self, ignore_converters=False):
        """Return a de-duplicated list of the keys we know about"""
        return self.storage.keys_after(self.prefix_string, ignore_converters=ignore_converters)
    reversed_keys = keys

    def items(self, ignore_converters=False):
        """Iterate over [(key, value), ...] pairs"""
        for key in self.keys(ignore_converters=ignore_converters):
            yield key, self.__getitem__(key, ignore_converters=ignore_converters)

    @versioned_iterable
    def values(self):
        """Return the values in the MergedOptions."""
        for key in self.keys():
            yield self[key]

    def __repr__(self):
        return "MergedOptions({0})".format(self.prefix_string)

    def path(self, path, **kwargs):
        """Return us a path instance"""
        for key, val in (("configuration", self), ("converters", self.converters), ("ignore_converters", self.ignore_converters)):
            if key not in kwargs:
                kwargs[key] = val
        return Path(path, **kwargs)

    def converted_path(self, path, ignore_converters=False, converters=None):
        """Convert a path into a Path object with a prefixed path"""
        if converters is None:
            converters = self.converters

        path_type = type(path)
        isPath = path_type is Path
        if isPath and path.ignore_converters is ignore_converters and path.converters is converters:
            return path

        if isPath:
            joined = path.joined()
        else:
            from option_merge import helper as hp
            if path_type in (list, tuple):
                path, joined = hp.prefixed_path_list(path, self.prefix_list)
            else:
                path, joined = hp.prefixed_path_string(path, self.prefix_string)

        return Path.convert(path, self, converters=converters, joined=joined).ignoring_converters(ignore_converters)

    def add_converter(self, converter):
        """Add a converter to our collection"""
        if converter not in self.converters:
            self.converters.append(converter)

    def install_converters(self, converters, make_converter):
        """
        For each specified converter, make a converter function and install
        a converter for that name.

        .. code-block:: python

            def make_converter(name, transformer):
                def convert(path, val):
                    return transformer(val)
                return convert

            m = MergedOptions.using({"a": 1, "b": 2})
            m.install_converters({"a": lambda v: v+1, "b": lambda v: v*2}, make_converter)
            m.converters.activate()

            assert m['a'] == 2
            assert m['b'] == 4
        """
        for name, spec in converters.items():
            convert = make_converter(name, spec)
            self.add_converter(Converter(convert=convert, convert_path=[name]))

    def as_dict(self, key="", ignore_converters=True, seen=None, ignore=None):
        """Collapse the storage at this prefix into a single dictionary"""
        return self.storage.as_dict(self.converted_path(key, ignore_converters=ignore_converters), seen=seen, ignore=ignore)




"""
Option Merge now gives you the ability to extend your program using "generic"
option_merge addons.

An Option Merge addon is a module with a ``__option_merge__`` hook and is
registered under the ``option_merge.addons`` pkg_resources entry_point.

Then, as the application developer, you must have ``input_algorithms>=0.5``
in your application's dependencies and you may use ``Collector#register_addons``
to register the addons you wish to use.

So, let's build an addon!::

    addon/
        my_option_merge_addon.py
        setup.py

my_option_merge_addon
    .. code-block:: python

        from input_algorithms import spec_base as sb

        def __option_merge__(configuration, result_maker, **kwargs):
            specs = {
                ((0, ("one", )): sb.integer_spec()
              }
            return result_maker(specs=specs, addons=[])

setup.py
    .. code-block:: python

        from setuptools import setup

        setup(
              name = "my_option_merge_addon"
            , version = "0.1"
            , packages = ['my_option_merge_addon']

            , entry_points =
              { "option_merge.addons":
                [ "mine = my_option_merge_addon"
                ]
              }
            )

Now as long as we have pip installed this addon, we can do something like:

.. code-block:: python

    from option_merge.collector import Collector
    from option_merge import MergedOptions
    from option_merge.addons import Addon

    from input_algorithms.meta import Meta

    configuration = MergedOptions.using({"one": "1"})
    Collector().register_addons(Addon, ["mine"], Meta, configuration)

    assert configuration["one"] == 1

Please see http://input-algorithms.readthedocs.io for more information about
how input_algorithms works.

Essentially your addon must use the passed in ``result_maker`` to return specs
to add to the configuration and additional ``addons`` to load.
"""
def find_input_algorithms(package="input_algorithms", version=">=0.5"):
    try:
        __import__(package)
    except ImportError:
        raise Exception("To use option_merge addons your program must also have {0} as a requirement".format(package))

    import pkg_resources
    try:
        pkg_resources.working_set.require(["{0}{1}".format(package, version)])
    except (pkg_resources.DistributionNotFound, pkg_resources.VersionConflict) as error:
        raise Exception("Have invalid version of {0}\terror={1}".format(package, error))

find_input_algorithms()

from input_algorithms.dictobj import dictobj
from input_algorithms import spec_base as sb
from input_algorithms.meta import Meta

import pkg_resources
import logging

log = logging.getLogger("option_merge.addons")

class Result(dictobj.Spec):
    specs = dictobj.Field(lambda: sb.dictof(sb.tupleof(sb.integer_spec(), sb.tuple_spec(sb.string_spec())), sb.has("normalise")))
    addons = dictobj.Field(sb.string_spec, wrapper=sb.listof)
    post_register = dictobj.Field(sb.any_spec, default=lambda *args, **kwargs:None)

class Addon(object):
    class NoSuchAddon(Exception):
        _fake_delfick_error = True
        def __init__(self, addon):
            self.message = ""
            self.addon = addon
            self.kwargs = dict(addon=addon)
        def __str__(self):
            return "No Such addon\taddon={0}".format(self.addon)

    class BadImport(Exception):
        _fake_delfick_error = True
        def __init__(self, message, importing, module, error):
            self.message = message
            self.error = error
            self.module = module
            self.importing = importing
            self.kwargs = dict(importing=importing, module=module, error=error)
        def __str__(self):
            return "Bad import: {0}\timporting={1}\tmodule={2}\terror={3}".format(self.message, self.importing, self.module, self.error)

    class BadAddon(Exception):
        _fake_delfick_error = True
        def __init__(self, message, expected, got, addon):
            self.message = message
            self.got = got
            self.expected = expected
            self.addon = addon
            self.kwargs = dict(expected=expected, got=got, addon=addon)
        def __str__(self):
            return "Bad addon: {0}\texpected={1}\tgot={2}\taddon={3}".format(self.message, self.expected, self.got, self.addon)

    @classmethod
    def get(kls, entry_point_name, configuration, **kwargs):
        entry_points = list(pkg_resources.iter_entry_points("option_merge.addons", name=entry_point_name))
        if len(entry_points) > 1:
            log.warning("Found multiple entry_points for option_merge.addons.{0}".format(entry_point_name))
        elif len(entry_points) == 0:
            raise kls.NoSuchAddon(addon="option_merge.addons.{0}".format(entry_point_name))
        else:
            log.info("Found {0} addon".format(entry_point_name))

        def result_maker(**data):
            return Result.FieldSpec().normalise(Meta(data, []), data)

        for entry_point in entry_points:
            try:
                module = entry_point.resolve()
            except ImportError as error:
                raise kls.BadImport("Error whilst resolving entry_point", importing=entry_point_name, module=entry_point.module_name, error=str(error))

            if not hasattr(module, "__option_merge__"):
                raise kls.BadImport("Error after resolving entry_point", importing=entry_point_name, module=entry_point.module_name, error="Didn't find an __option_merge__ function")

            r = module.__option_merge__(configuration, result_maker, **kwargs)
            if not issubclass(r.__class__, Result):
                raise kls.BadAddon("Addon's __option_merge__ returned something wrong!", expected=Result, got=r.__class__, addon=entry_point_name)

            yield r




from option_merge.merge import ConverterProperty, KeyValuePairsConverter, AttributesConverter
from option_merge.merge import MergedOptions
from option_merge.converter import Converter
from option_merge.not_found import NotFound

# Explicitly make these part of this module
NotFound = NotFound
Converter = Converter
MergedOptions = MergedOptions
ConverterProperty = ConverterProperty
AttributesConverter = AttributesConverter
KeyValuePairsConverter = KeyValuePairsConverter




"""
Helpers for joining together things
"""

import six

Path = None
list_types = (list, tuple)
string_types = (str, ) + six.string_types

def dot_joiner(item, item_type=None):
    """Join lists of list of strings with a single dot in between each"""
    if item_type in string_types:
        return item
    else:
        return dot_join_item(item, item_type or type(item))

def dot_join_item(item, item_type):
    """
    Join list of strings with a single dot in between each

    Note that if the items have dots in them, then it's possible to have multiple dots.

    This is fine, as long as we are consistent
    """
    if item_type not in list_types:
        return str(item)

    result = []
    for part in item:
        part_type = type(part)
        global Path
        if Path is None:
            from option_merge.path import Path

        if part_type is Path:
            joined = part.joined()
            if joined:
                result.append(joined)
            continue

        if part_type in list_types:
            part = ''.join(part)

        if part:
            result.append(part)

    return '.'.join(str(part) for part in result)

def join(one, two):
    """
    Join two paths together

    Where either path is either string, list of strings or Path
    """
    global Path
    if Path is None:
        from option_merge.path import Path

    if isinstance(one, Path):
        one = one.path
    if isinstance(two, Path):
        two = two.path

    if isinstance(one, six.string_types):
        if isinstance(two, six.string_types):
            joined = [one] + [two]
        else:
            joined = [one] + two
    else:
        if isinstance(two, six.string_types):
            joined = one + [two]
        else:
            joined = one + two

    return [item for item in joined if item]




"""
MergedOptionStringFormatter is provided as a base class for creating a special
formatter that can format strings using a MergedOptions object.

.. code-block:: python

    from delfick_error import DelfickError

    class BadOptionFormat(DelfickError): pass

    class MyFormatter(MergedOptionStringFormatter):
        def get_string(self, key):
            if key not in self.all_options:
                kwargs = {}
                if len(self.chain) > 1:
                    kwargs['source'] = Meta(self.all_options, self.chain[-2]).source
                raise BadOptionFormat("Can't find key in options", key=key, chain=self.chain, **kwargs)

            return super(MyFormatter, self).get_string(key)

        def special_get_field(self, value, args, kwargs, format_spec=None):
            if format_spec in ("env", ):
                return value, ()

            if value in self.chain:
                raise BadOptionFormat("Recursive option", chain=self.chain + [value])

        def special_format_field(self, obj, format_spec):
            if format_spec == "env":
                return "${{{0}}}".format(obj)

    m = MergedOptions.using({"a": {"b": 3}, "c": 5})
    formatted = MyFormatter(m, "", "a.b: {a.b} and c={c}").format()
    assert formatted == "a.b: 3 and c=5"
"""

from option_merge.versioning import VersionedDict
from option_merge.merge import MergedOptions

import string
import types
import six

class NotSpecified(object):
    """The difference between None and not specified"""

class MergedOptionStringFormatter(string.Formatter):
    """
    Resolve format options into a MergedOptions dictionary
    """
    def __init__(self, all_options, option_path, chain=None, value=NotSpecified):
        if chain is None:
            if isinstance(option_path, list):
                chain = [thing for thing in option_path]
            else:
                chain = [option_path]
        self.chain = chain
        self.value = value
        self.option_path = option_path
        self.all_options = all_options
        super(MergedOptionStringFormatter, self).__init__()

    def format(self):
        """Format our option_path into all_options"""
        val = self.value
        if val is NotSpecified:
            val = self.get_string(self.option_path)

        if not isinstance(val, six.string_types):
            return val
        return super(MergedOptionStringFormatter, self).format(val)

    def special_get_field(self, value, args, kwargs, format_spec=None):
        """
        Must be implemented

        In this function it is recommended you do something similar to:

        .. code-block:: python

            if value in self.chain:
                raise ValueError("Recursive option")

        But raise a more meaningful error!

        You may also return (value, ()) if ``format_spec`` is a custom format.
        """
        raise NotImplementedError()

    def special_format_field(self, obj, format_spec):
        """
        Must be implemented

        In this function you match against ``format_spec`` and return either a
        formatted version of ``obj`` or None.

        .. code-block:: python

            class MyFormatter(MergedOptionStringFormatter):
                def special_get_field(self, value, args, kwargs, format_spec=None):
                    if format_spec == "plus_one":
                        return value, ()

                def special_format_field(obj, format_spec):
                    if format_spec == "plus_one":
                        return int(obj) + 1

            m = MergedOptions.using({"a": 3})
            formatted = MyFormatter(m, "", "{a:plus_one}").format()
            assert formatted == "4"
        """
        raise NotImplementedError()

    def with_option_path(self, value):
        """Clone this instance with the new value as option_path and no override value"""
        return self.__class__(self.all_options, value, chain=self.chain + [value], value=NotSpecified)

    def get_string(self, key):
        """
        Get a string from all_options

        it is recommended you override this method and raise an error if the key
        does not exist in self.all_options..
        """
        return self.all_options[key]

    def get_field(self, value, args, kwargs, format_spec=None):
        """Also take the spec into account"""
        special = self.special_get_field(value, args, kwargs, format_spec)
        if special is not None:
            return special
        else:
            return self.with_option_path(value).format(), ()

    def format_field(self, obj, format_spec):
        """Know about any special formats"""
        special = self.special_format_field(obj, format_spec)
        if special:
            return special
        else:
            is_dict = type(obj) in (VersionedDict, MergedOptions) or isinstance(obj, dict)
            is_a_mock = hasattr(obj, 'mock_calls')
            is_special_type = any(isinstance(obj, typ) for typ in (types.LambdaType, types.FunctionType, types.MethodType, types.BuiltinFunctionType, types.BuiltinMethodType))

            if is_dict or is_special_type or is_a_mock:
                return obj
            else:
                return super(MergedOptionStringFormatter, self).format_field(obj, format_spec)

    def vformat(self, format_string, args, kwargs):
        """This changes in 3.5.1 and I want it to not have changed"""
        used_args = set()
        result = self._vformat(format_string, args, kwargs, used_args, 2)
        self.check_unused_args(used_args, args, kwargs)
        return result

    def _vformat(self, format_string, args, kwargs, used_args, recursion_depth):
        """I really want to know what the format_string is so I'm taking from standard library string and modifying slightly"""
        if recursion_depth < 0:
            raise ValueError('Max string recursion exceeded')

        result = []

        for literal_text, field_name, format_spec, conversion in self.parse(format_string):

            # output the literal text
            if literal_text:
                result.append(literal_text)

            # if there's a field, output it
            if field_name is not None:
                # this is some markup, find the object and do
                #  the formatting

                # given the field_name, find the object it references
                #  and the argument it came from
                # Slight modification here to pass in the format_spec
                obj, arg_used = self.get_field(field_name, args, kwargs, format_spec)
                used_args.add(arg_used)

                # do any conversion on the resulting object
                obj = self.convert_field(obj, conversion)

                # expand the format spec, if needed
                format_spec = self._vformat(format_spec, args, kwargs,
                                            used_args, recursion_depth-1)

                # format the object and append to the result
                result.append(self.format_field(obj, format_spec))

        if len(result) == 1:
            return result[0]
        return ''.join(str(obj) for obj in result)




"""
The Storage class is used to encapsulate the underlying data structure for
MergedOptions.

It's main functionality is the ``get_info`` method which yields DataPath objects
which can be used to get keys and values for some path.

It is also used to get thesource for particular paths.
"""

from option_merge.versioning import versioned_iterable, versioned_value, VersionedDict
from option_merge.merge import MergedOptions
from option_merge.not_found import NotFound
from option_merge.value_at import value_at
from option_merge.joiner import dot_joiner
from option_merge import helper as hp
from option_merge.path import Path

from delfick_error import ProgrammerError

class DataPath(object):
    """
    Encapsulates a (path, data, source) triplet and getting keys and values from
    that path in the data.
    """

    def __init__(self, path, data, source=None):
        self.path = path
        self.data = data
        self.source = source

    @property
    def is_dict(self):
        is_dict = getattr(self, "_is_dict", None)
        if is_dict is None:
            is_dict = self._is_dict = type(self.data) in (dict, VersionedDict, MergedOptions) or isinstance(self.data, dict)
        return is_dict

    def items(self, prefix, want_one=False):
        """
        Yield (key, data, short_path) triplets for after this prefix
        We don't return values to avoid any lazy computations
        Where short_path is the path (not including the key) from data

        First we remove the prefix from the path.

        If there is more prefix than path
            Dig into the data

        If there is more path than prefix
            yield the next part of the pa

        If the prefix isn't in the path
            raise NotFound

        If the prefix isn't in combination of path and data
            raise NotFound
        """
        data = self.data
        joined = self.path.joined()
        joined_prefix = prefix.joined()

        if not prefix or joined == joined_prefix or joined.startswith(joined_prefix + '.'):
            shortened_path = self.path.without(prefix)
            if shortened_path:
                yield list(shortened_path)[0], data, list(shortened_path)[1:]
                return
            else:
                prefix = ""
        else:
            if not joined_prefix.startswith(joined):
                raise NotFound
            else:
                prefix = Path(prefix).without(self.path)

        while True:
            if not prefix and want_one:
                yield "", data, []
                return

            if type(data) not in (dict, VersionedDict, MergedOptions) and getattr(data, "is_dict", False) is not True:
                raise NotFound

            if not prefix:
                for key in data.keys():
                    yield key, data, []
                return

            found = False
            for key in reversed(sorted(data.keys(), key=len)):
                if prefix.startswith(key):
                    data = data[key]
                    prefix = prefix.without(key)
                    found = True
                    break

            if not found:
                raise NotFound

    def keys_after(self, prefix):
        """Yield the keys after this prefix"""
        for key, _, _ in self.items(prefix):
            yield key

    @versioned_value
    def value_after(self, prefix):
        """Returns the value after prefix"""
        for key, data, short_path in self.items(prefix, want_one=True):
            if short_path:
                return hp.make_dict(key, short_path, data)
            else:
                if key:
                    return hp.make_dict(key, short_path, data)
                else:
                    return data

class Storage(object):
    """
    Holds the data used by MergedOptions.

    It understands the different sources of data that makes up the whole, how to
    get information for particular paths, how to delete particular paths, and
    how to get the sources for particular paths.
    """

    def __init__(self):
        self.data = []
        self.deleted = []
        self._version = -1

    ########################
    ###   USAGE
    ########################

    def add(self, path, data, source=None):
        """Add data at the beginning"""
        if not isinstance(path, Path):
            raise ProgrammerError("Path should be a Path object\tgot={0}".format(type(path)))
        self._version += 1
        self.data.insert(0, (path, VersionedDict.convert(data), source))

    def get(self, path):
        """Get a single value from a path"""
        for info in self.get_info(path):
            return info.data
        raise KeyError(path)

    def source_for(self, path, chain=None):
        """Find all the sources for a given path"""
        sources = []
        if chain is None:
            chain = []
        if (path, self) in chain:
            return []
        if not path:
            return []

        for info in self.get_info(path, ignore_converters=True, chain=chain + [(path, self)]):
            source = info.source
            if callable(info.source):
                source = info.source()

            if info.path == path and not info.is_dict:
                if isinstance(source, list):
                    return list(source)
                elif source:
                    return [source]
            else:
                if source not in sources:
                    if source:
                        if isinstance(source, list):
                            sources.extend(source)
                        else:
                            sources.append(source)

        if len(sources) == 1:
            return sources[0]
        else:
            return sources

    def delete(self, path):
        """Delete the first instance of some path"""
        for index, (info_path, data, _) in enumerate(self.data):
            dotted_info_path = dot_joiner(info_path)
            if dotted_info_path == path:
                self._version += 1
                del self.data[index]
                return
            elif dotted_info_path.startswith("{0}.".format(path)):
                self._version += 1
                del self.data[index]
                return
            elif not dotted_info_path or path.startswith("{0}.".format(dotted_info_path)):
                remainder = path
                if info_path:
                    remainder = Path.convert(path).without(dotted_info_path)
                if self.delete_from_data(data, remainder):
                    return

        raise KeyError(path)

    ########################
    ###   IMPLEMENTATION
    ########################

    @property
    def version(self):
        if self._version > 0:
            return [self._version] + [getattr(item, "version", 0) for item in self.data]
        else:
            return -1

    @versioned_iterable
    def get_info(self, path, ignore_converters=False, chain=None):
        """Yield DataPath objects for this path in the data"""
        yielded = False
        if not self.data:
            return

        if chain is None:
            chain = []

        ignore_converters = ignore_converters or getattr(path, 'ignore_converters', False)
        path = Path.convert(path).ignoring_converters(ignore_converters)

        for info_path, data, source in self.data:
            for full_path, found_path, val in self.determine_path_and_val(path, info_path, data, source):
                source = self.make_source_for_function(val, found_path, chain, default=source)
                yield DataPath(full_path, val, source)
                yielded = True

        if not yielded:
            raise KeyError(path)

    def determine_path_and_val(self, path, info_path, data, source):
        """
        Yield the full_path, found_path and val for this path into this data and info_path

        Where found_path is the path relative to the data
        """
        joined_path = dot_joiner(path)
        joined_info_path = dot_joiner(info_path)
        if joined_path == joined_info_path:
            yield info_path, "", data
            return

        try:
            if not info_path:
                found_path, val = value_at(data, Path.convert(path), self)
                yield info_path + found_path, dot_joiner(found_path, list), val
                return

            if joined_path.startswith(joined_info_path + '.'):
                get_at = Path.convert(path).without(info_path)
                found_path, val = value_at(data, get_at, self)
                yield info_path + found_path, dot_joiner(found_path), val
                return

            # We are only part way into info_path
            path = Path.convert(path)
            for key, data, short_path in DataPath(Path.convert(info_path), data, source).items(path, want_one=True):
                yield path, "", hp.make_dict(key, short_path, data)
        except NotFound:
            pass

    def make_source_for_function(self, obj, path, chain, default=None):
        """Return us a function that will get the source for some path on the specified obj"""
        def source_for():
            if hasattr(obj, "source_for"):
                nxt = obj.source_for(path, chain)
                if nxt:
                    return nxt
            return default
        return source_for

    @versioned_iterable
    def keys_after(self, path, ignore_converters=False):
        """Get all the keys after this path"""
        done = set()
        stopped = set()
        for info in self.get_info(path, ignore_converters=ignore_converters):
            if hasattr(info.data, "storage") and info.data.storage is self:
                continue

            try:
                for key in info.keys_after(Path.convert(path)):
                    if path:
                        joined = dot_joiner([path, key], list)
                    else:
                        joined = key

                    if not any(s == joined or joined.startswith("{0}.".format(s)) for s in stopped):
                        if key not in done:
                            yield key
                            done.add(key)
            except NotFound:
                pass

            if not info.is_dict:
                stopped.add(dot_joiner(info.path))

    def delete_from_data(self, data, path):
        """Delete this path from the data"""
        if not path or (type(data) not in (dict, VersionedDict, MergedOptions) and not isinstance(data, dict)):
            return False

        keys = list(reversed(sorted(data.keys())))
        if path in keys:
            self._version += 1
            del data[path]
            return True

        for key in keys:
            if path.startswith("{0}.".format(key)):
                if self.delete_from_data(data[key], Path.convert(path).without(key)):
                    return True

    def as_dict(self, path, seen=None, ignore=None):
        """Return this path as a single dictionary"""
        result = {}
        if seen is None:
            seen = {}

        if path not in seen:
            seen[path] = []
        if self in seen[path]:
            return {}
        seen[path].append(self)

        for i in range(len(self.data)-1, -1, -1):
            prefix, data, _ = self.data[i]

            if prefix:
                prefixer = list(prefix)
                while prefixer:
                    key = prefixer.pop()
                    data = VersionedDict({key: data})

            val = data
            used = None
            found = False
            path_without_prefix = path
            while not found or path_without_prefix:
                if hasattr(data, "is_dict") and data.is_dict:
                    if hasattr(val, "as_dict"):
                        val = val.as_dict(path_without_prefix, seen=seen, ignore=ignore)
                        path_without_prefix = ""
                    else:
                        val = dict(val)

                try:
                    used, val = value_at(val, path_without_prefix, self)
                    found = True
                except NotFound:
                    found = False
                    break

                if used and path_without_prefix:
                    path_without_prefix = path_without_prefix.without(dot_joiner(used))

            if found and path_without_prefix == "":
                is_dict = lambda item: type(item) in (dict, VersionedDict, MergedOptions) or isinstance(item, dict)
                if not is_dict(val):
                    result = val
                else:
                    if not is_dict(result):
                        result = {}
                    hp.merge_into_dict(result, val, seen, ignore=ignore)

        return result




"""
The collector object is responsible for collecting configuration
and setting up converters

For example:

.. code-block:: python

    class JsonCollector(Collector):
        def start_configuration(self):
            return MergedOptions()

        def find_missing_config(self, configuration):
            assert "important_option" in configuration

        def extra_prepare(self, configuration, args_dict):
            configuration.update(
                  {"some_option": args_dict["some_option"]}
                , source=<extra_prepare>
                )

        def read_file(self, location):
            return json.loads(location)

        def add_configuration(self, configuration, collect_another_source, done, result, src):
            configuration.update(result, source=src)
            for loc in result.get("others", []):
                collect_another_source(
                      os.path.join(os.path.dirname(src), loc)
                    , prefix=os.path.splitext(loc)[1]
                    )

        def extra_configuration_collection(self, configuration):
            def convert(p, v):
                return v * 2
            configuration.add_converter(
                  Converter(
                      convert=convert
                    , convert_path=["some", "contrived", "example"]
                    )
                )

    collector = JsonCollector()
    collector.prepare(
          "/path/to/first_config_file.json"
        , {"some_option": "stuff"}
        )

    #### /path/to/first_config_file.json
    # { "hello": "there"
    # , "others": ["some.json"]
    # }

    #### /path/to/some.json
    # { "contrived":
    #   { "example": 2
    #   }
    # }

    collector.configuration["some_option"] == "stuff"
    collector.configuration["some.contrived.example"] == 4
"""

from option_merge.versioning import VersionedDict
from option_merge.converter import Converter

from getpass import getpass
import logging
import os

log = logging.getLogger("option_merge.collector")

class Collector(object):
    """
    When using the Collector, it is expected that you implement a number of hooks
    to make this class useful.
    """
    class BadFileErrorKls(Exception):
        _fake_delfick_error = True
        def __init__(self, message):
            self.message = message
            self.kwargs = {}
        def __str__(self):
            return "BadFile: {0}".format(self.message)

    class BadConfigurationErrorKls(Exception):
        _fake_delfick_error = True
        def __init__(self, _errors):
            self.message = ""
            self.kwargs = {}
            self.errors = _errors
        def __str__(self):
            message = "errors:\n=======\n\n\t{0}".format("\n\t".join("{0}\n-------".format('\n\t'.join(str(error).split('\n'))) for error in self.errors))
            return "BadConfiguration:\n{0}".format(message)

    def __init__(self):
        self.setup()

    ########################
    ###   HOOKS
    ########################

    def setup(self):
        """Called at __init__ time"""

    def alter_clone_args_dict(self, new_collector, new_args_dict, *args, **kwargs):
        """
        Hook for altering args_dict given to a clone collector it must return a dictionary

        This dictionary will be used in the ``prepare`` call for the new collector
        """
        return new_args_dict

    def find_missing_config(self, configuration):
        """Hook to raise errors about missing configuration"""

    def extra_prepare(self, configuration, args_dict):
        """Hook for any extra preparation before the converters are activated"""

    def extra_prepare_after_activation(self, configuration, args_dict):
        """Hook for any extra preparation after the converters are activated"""

    def home_dir_configuration_location(self):
        """Hook to return the location of the configuration in the user's home directory"""
        return None

    def read_file(self, location):
        """Hook to read in a file and return a dictionary"""
        raise NotImplementedError()

    def start_configuration(self):
        """Hook for starting the base of the configuration"""
        raise NotImplementedError()

    def add_configuration(self, configuration, collect_another_source, done, result, src):
        """
        Hook to add to the configuration the loaded result from src into configuration

        The collect_another_source function can be used to collect another source

        And done is a dictionary of configuration we have already added
        """
        raise NotImplementedError()

    def extra_configuration_collection(self, configuration):
        """Hook to do any extra configuration collection or converter registration"""

    ########################
    ###   USAGE
    ########################

    def clone(self, *args, **kwargs):
        """Create a new collector that is a clone of this one"""
        if not hasattr(self, "configuration_file"):
            return self.__class__()
        new_collector = self.__class__()
        args_dict_clone = dict(self.configuration["args_dict"].items())
        new_args_dict = self.alter_clone_args_dict(new_collector, args_dict_clone, *args, **kwargs)
        new_collector.prepare(self.configuration_file, new_args_dict)
        return new_collector

    def prepare(self, configuration_file, args_dict):
        """
        Prepare the collector!

        * Collect all the configuration
        * find missing configuration
        * do self.extra_prepare
        * Activate the converters
        * Do self.extra_prepare_after_activation
        """
        self.configuration_file = configuration_file
        self.configuration = self.collect_configuration(configuration_file, args_dict)

        self.find_missing_config(self.configuration)

        self.extra_prepare(self.configuration, args_dict)
        self.configuration.converters.activate()
        self.extra_prepare_after_activation(self.configuration, args_dict)

        for addon, result, kwargs in getattr(self, "registered_addons", []):
            result.post_register(self.configuration, **kwargs)

    def register_addons(self, AddonGetter, addons, Meta, configuration, **kwargs):
        """
        Resolve and add addons into the configuration.

        Addons should be a list of strings to entry_points at option_merge.addons
        """
        self.registered_addons = getattr(self, "registered_addons", [])

        def register(adns):
            moar = []
            for addon in adns:
                if addon in [name for name in self.registered_addons]:
                    continue

                for result in AddonGetter.get(addon, configuration, **kwargs):
                    self.registered_addons.insert(0, (addon, result, kwargs))
                    self.register_converters(result.specs, Meta, configuration)
                    moar.extend(result.addons)
            return moar

        nxt = addons
        while True:
            nxt = register(nxt)
            if not nxt:
                break

    def register_converters(self, specs, Meta, configuration):
        """
        Register converters

        specs
            a Dictionary of {(priority, [path]): spec}

        Meta
            The class to instantiate to create ``meta``

            The converter will call ``spec.normalise(meta, val)``

        configuration
            The configuration to add the converter to
        """
        for (_, key), spec in sorted(specs.items()):
            def make_converter(k, s):
                def converter(p, v):
                    log.info("Converting %s", p)
                    meta = Meta(p.configuration, [])
                    for kk in k:
                        meta = meta.at(kk)
                    configuration.converters.started(p)
                    return s.normalise(meta, v)
                configuration.add_converter(Converter(convert=converter, convert_path=k))
            make_converter(key, spec)

    ########################
    ###   CONFIG
    ########################

    def collect_configuration(self, configuration_file, args_dict):
        """Return us a MergedOptions with this configuration and any collected configurations"""
        errors = []

        configuration = self.start_configuration()

        configuration.update(
              { "getpass": getpass
              , "collector": self
              , "args_dict": args_dict
              }
            , source = "<preparation>"
            )

        sources = []
        if configuration_file:
            sources.append(configuration_file)

        home_dir_configuration = self.home_dir_configuration_location()
        if home_dir_configuration:
            sources.insert(0, home_dir_configuration)

        done = set()
        def add_configuration(src, prefix=None, extra=None):
            log.info("Adding configuration from %s", os.path.abspath(src))
            if os.path.abspath(src) in done:
                return
            else:
                done.add(os.path.abspath(src))

            if src is None or not os.path.exists(src):
                return

            try:
                if os.stat(src).st_size == 0:
                    result = VersionedDict({})
                else:
                    result = VersionedDict(self.read_file(src))
            except self.BadFileErrorKls as error:
                errors.append(error)
                return

            if not result:
                return

            if extra:
                result.update(extra)
            result["config_root"] = os.path.abspath(os.path.dirname(src))

            while prefix:
                part = prefix.pop()
                result = VersionedDict({part: result})

            self.add_configuration(configuration, add_configuration, done, result, src)

        for source in sources:
            add_configuration(source)

        self.extra_configuration_collection(configuration)

        if errors:
            raise self.BadConfigurationErrorKls("Some of the configuration was broken", _errors=errors)

        return configuration




import random
import time

class versioned_value(object):
    """
    A property that holds a cache of {prefix: {ignore_converters: value}}

    Where the entire cache is revoked if instance.version changes number
    """
    class NotYet(object): pass
    class First(object): pass

    def __init__(self, func):
        self.func = func
        self.cached_key = "_{0}_cached".format(self.func.__name__)
        self.expected_version_key = "_{0}_expected_version".format(self.func.__name__)

    def __get__(self, instance=None, owner=None):
        def returned(*args, **kwargs):
            version = getattr(instance, "version", 0)
            if version is -1:
                return self.func(instance, *args, **kwargs)

            if args:
                prefix = str(args[0])
            else:
                prefix = getattr(instance, "prefix_string", "")

            # Ignore_converters can be specified in three places, in this order
            # kwarg to the function
            # property on the path
            # property on the instance
            ignore_converters = kwargs.get('ignore_converters', getattr(prefix, 'ignore_converters', getattr(instance, 'ignore_converters', False)))

            cached = getattr(instance, self.cached_key, {})
            expected_version = getattr(instance, self.expected_version_key, self.First)

            if version != expected_version:
                cached = {}
                expected_version = version

            # We set the caches on the instance
            # so that they are garbage collected
            # When the instance gets deleted
            setattr(instance, self.cached_key, cached)
            setattr(instance, self.expected_version_key, expected_version)

            if cached.get(prefix, {}).get(ignore_converters, self.NotYet) is self.NotYet:
                if prefix not in cached:
                    cached[prefix] = {}
                try:
                    cached[prefix][ignore_converters] = (self.func(instance, *args, **kwargs), False)
                except KeyError as error:
                    cached[prefix][ignore_converters] = (error, True)

            val, is_error = cached[prefix][ignore_converters]
            if is_error:
                raise val
            else:
                return val
        return returned

class versioned_iterable(object):
    """
    A property that holds a cache of {instance: {prefix: {ignore_converters: iterator}}}

    Where the entire cache is revoked if instance.version changes number
    """
    class First(object): pass
    class NotYet(object): pass
    class Finished(object): pass

    def __init__(self, func):
        self.func = func
        self.cached_key = "_{0}_cached".format(self.func.__name__)
        self.value_cache_key = "_{0}_value_cache".format(self.func.__name__)
        self.expected_version_key = "_{0}_expected_version".format(self.func.__name__)

    def iterator_for(self, instance, prefix, ignore_converters, cached, value_cache):
        for item in value_cache[prefix][ignore_converters]:
            yield item

        iterator = cached[prefix][ignore_converters]
        if iterator is self.Finished:
            return
        ident = cached[prefix][ignore_converters] = "{0}:{1}".format(time.time(), random.random())

        try:
            while True:
                nxt = next(iterator)
                if cached.get(prefix, {}).get(ignore_converters) == ident:
                    value_cache[prefix][ignore_converters].append(nxt)
                yield nxt
        except StopIteration:
            iterator.close()
            if cached.get(prefix, {}).get(ignore_converters) == ident:
                cached[prefix][ignore_converters] = self.Finished

    def __get__(self, instance=None, owner=None):
        def returned(*args, **kwargs):

            version = getattr(instance, "version", 0)
            if version is -1:
                return self.func(instance, *args, **kwargs)

            if args:
                prefix = args[0]
            else:
                prefix = getattr(instance, "prefix_string", "")

            # Ignore_converters can be specified in three places, in this order
            # kwarg to the function
            # property on the path
            # property on the instance
            ignore_converters = kwargs.get('ignore_converters', getattr(prefix, 'ignore_converters', getattr(instance, 'ignore_converters', False)))

            cached = getattr(instance, self.cached_key, {})
            value_cache = getattr(instance, self.value_cache_key, {})
            expected_version = getattr(instance, self.expected_version_key, self.First)

            if version != expected_version:
                cached = {}
                value_cache = {}
                expected_version = version

            # We set the caches on the instance
            # so that they are garbage collected
            # When the instance gets deleted
            setattr(instance, self.cached_key, cached)
            setattr(instance, self.value_cache_key, value_cache)
            setattr(instance, self.expected_version_key, expected_version)

            if cached.get(prefix, {}).get(ignore_converters, self.NotYet) is not self.Finished:
                if prefix not in cached:
                    cached[prefix] = {}
                if prefix not in value_cache:
                    value_cache[prefix] = {}

                value_cache[prefix][ignore_converters] = []
                ret = self.func(instance, *args, **kwargs)
                if isinstance(ret, list):
                    value_cache[prefix][ignore_converters] = ret
                    cached[prefix][ignore_converters] = self.Finished
                else:
                    cached[prefix][ignore_converters] = ret
            return self.iterator_for(instance, prefix, ignore_converters, cached, value_cache)
        return returned

class VersionedDict(object):
    """
    A wrapper for dictionaries that has a version property

    The version starts at 0 and is incremented whenever __setitem__ or __delitem__ is used
    """
    @classmethod
    def convert(kls, item):
        from option_merge import MergedOptions
        if isinstance(item, dict):
            if not isinstance(item, VersionedDict) and not isinstance(item, MergedOptions):
                return kls(item)
        return item

    def __init__(self, data):
        self.data = data
        self.version = 0
        self.is_dict = True
        super(VersionedDict, self).__init__()

    def __setitem__(self, key, val):
        self.version += 1
        self.data[key] = val

    def __delitem__(self, key):
        self.version += 1
        del self.data[key]

    def __eq__(self, other):
        return self.data == other

    def __nonzero__(self):
        return True if self.data else False

    def __bool__(self):
        return bool(self.data)

    def keys(self):
        return self.data.keys()

    @versioned_value
    def reversed_keys(self):
        return list(reversed(sorted(self.keys(), key=lambda d: len(str(d)))))

    def values(self):
        return self.data.values()

    def update(self, other):
        self.data.update(other)

    def items(self):
        return self.data.items()

    def get(self, key, dflt=None):
        return self.data.get(key, dflt)

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, key):
        return self.data[key]

    def __repr__(self):
        return repr(self.data)

    def __str__(self):
        return self.data.__str__()

    def __unicode__(self):
        return self.data.__unicode__()




"""
Option_merge includes functionality to convert values on access.

This functionality is assisted by the Converter and Converters classes.

Converter is used to encapsulate a single converter, and converters is used to
group together multiple converters.

Usage is through the MergedOptions object, which has an instance of
Converters on it.

.. code-block:: python

    m = MergedOptions.using({"a": 1})
    converter = Converter(convert=lambda p, v: v+1, convert_path=['a'])
    m.add_converter(converter)

    assert m['a'] == 1
    m.converters.activate()
    assert m['a'] == 2
    assert m['a'] == 2

Note that the converters must be activated before use and subsequent calls to
converted paths will use a cached result.
"""

from option_merge.versioning import versioned_value
from option_merge.joiner import dot_joiner

class Converter(object):
    """
    Encapsulates a single converter.

    It contains the converting logic as well as a list of paths that this
    conversion should be used against.

    It has a method "matches" that is used against each possible path and will
    check for exact matches against the ``convert_path``.
    """
    def __init__(self, convert, convert_path=None):
        self.convert = convert
        self.convert_path = convert_path
        self.convert_path_joined = dot_joiner(convert_path)

    def __call__(self, path, data):
        """Proxy the conversion logic in ``self.convert``"""
        return self.convert(path, data)

    def matches(self, path):
        """Check to see if this converter should be used against this path"""
        if hasattr(path, "joined"):
            joined_path = path.joined()
        else:
            joined_path = dot_joiner(path)
        return self.convert_path and joined_path == self.convert_path_joined

class Converters(object):
    """
    Holds a group of converters.

    Has logic to say whether the converters are activated.

    Also memoizes the results of conversion.
    """
    def __init__(self):
        self._waiting = {}
        self._converted = {}
        self._converters = []
        self.version = 0
        self.activated = False

    def __iter__(self):
        """
        Iterate through the converters

        Yield nothing if not activated yet
        """
        if not self.activated:
            return iter([])
        return iter(self._converters)

    def append(self, converter):
        """Add a converter, we store these as a list"""
        self._converters.append(converter)
        self.version += 1

    def activate(self):
        """Mark the converters as activated"""
        self.activated = True
        self.version += 1

    @versioned_value
    def matches(self, path):
        """
        Return (converter, matches) where converter is the matched converter or None

        And matches is a boolean signifying whether there was a match
        """
        for converter in self:
            if not hasattr(converter, "matches") or converter.matches(path):
                return converter, True

        return None, False
    matches.debug = True

    def converted(self, path):
        """Return whether this path has been converted yet"""
        if not self.activated:
            return False
        return path in self._converted

    def converted_val(self, path):
        """
        Return the converted value for this path

        This function should be guarded via the use of ``self.converted``
        """
        return self._converted[path]

    def waiting(self, path):
        """Return whether we're waiting for this path"""
        return path in self._waiting

    def done(self, path, value):
        """Mark a path as been replaced by the specified value"""
        if path in self._waiting:
            del self._waiting[path]
        self._converted[path] = value

    def started(self, path):
        """Mark this path as waiting"""
        self._waiting[path] = True




from setuptools import setup, find_packages

setup(
      name = "option_merge"
    , version = "1.3"
    , packages = ['option_merge'] + ['option_merge.%s' % pkg for pkg in find_packages('option_merge')]
    , include_package_data = True

    , install_requires =
      [ 'six'
      ]

    , extras_require =
      { "tests":
        [ "noseOfYeti>=1.4.9"
        , "nose"
        , "mock"
        , 'delfick_error==1.7.3'
        ]
      }

    # metadata for upload to PyPI
    , url = "http://github.com/delfick/option_merge"
    , author = "Stephen Moore"
    , author_email = "stephen@delfick.com"
    , description = "Code to deeply merge multiple python dictionaries"
    , license = "WTFPL"
    , keywords = "deep merge"
    )



from option_merge.versioning import VersionedDict
from option_merge.merge import MergedOptions
from option_merge.not_found import NotFound
from option_merge.path import Path

def value_at(data, path, called_from=None, chain=None):
    """
    Return the value at this path

    It is assumed path is a Path object
    """
    if not chain:
        chain = []

    if not path:
        return chain, data

    data_type = type(data)
    if data_type not in (dict, VersionedDict, MergedOptions) and getattr(data, "is_dict", False) is not True:
        raise NotFound

    joined = path.joined()
    isMergedOptions = data_type is MergedOptions

    if not data:
        keys = []
    else:
        if hasattr(data, "reversed_keys"):
            keys = list(data.reversed_keys())
        else:
            keys = list(reversed(sorted(data.keys(), key=lambda d: len(str(d)))))

    if joined in keys:
        if isMergedOptions:
            da = data.get(path.path, ignore_converters=getattr(path, "ignore_converters", False))
        else:
            da = data[joined]

        if not chain:
            return path, da
        else:
            return chain + [path], da

    for key in keys:
        if path.first_part_is(key):
            try:
                if isMergedOptions:
                    nxt = data.get(key, ignore_converters=path.ignore_converters)
                else:
                    nxt = data[key]

                storage = getattr(nxt, "storage", None)
                if storage and called_from is storage:
                    raise NotFound

                prefix = path.without(key)
                if not prefix:
                    return chain+[key], nxt

                prefix = Path.convert(prefix).ignoring_converters(path.ignore_converters)
                return value_at(nxt, prefix, called_from, chain=chain+[key])
            except NotFound:
                pass

    raise NotFound

